#ifndef __IKD_OX_BASIC_OXFLOAT_H__
#define __IKD_OX_BASIC_OXFLOAT_H__

#include <math.h>
#include <stdint.h>
#include <string.h>

namespace OX {
	/**
	* @brief xNg(3)
	*
	* (x,y,z)3xNgł
	*/
	class Vec3 {
	public:
		union {
			struct { float x, y, z; };
			float f[ 3 ];
		};

	public:
		Vec3( float x = 0.0f, float y = 0.0f, float z = 0.0f ) : x( x ), y( y ), z( z ) {}

		//!< +Zq
		Vec3 operator +( const Vec3 &r ) const { return Vec3( x + r.x, y + r.y, z + r.z ); }
		//!< -Zq
		Vec3 operator -( const Vec3 &r ) const { return Vec3( x - r.x, y - r.y, z - r.z ); }
		//!< *Zq
		Vec3 operator *( const float &r ) const { return Vec3( x * r, y * r, z * r ); }
		//!< *Zq
		Vec3 operator *( const Vec3 &r ) const { return Vec3( x * r.x, y * r.y, z * r.z ); }
		//!< /Zq
		Vec3 operator /( const Vec3 &r ) const { return Vec3( x / r.x, y / r.y, z / r.z ); }
		//!< +=Zq
		Vec3 &operator +=( const Vec3 &r ) { x += r.x; y += r.y; z += r.z; return *this; }
		//!< -=Zq
		Vec3 &operator -=( const Vec3 &r ) { x -= r.x; y -= r.y; z -= r.z; return *this; }
		//!< *=Zq
		Vec3 &operator *=( const Vec3 &r ) { x *= r.x; y *= r.y; z *= r.z; return *this; }
		//!< /=Zq
		Vec3 &operator /=( const Vec3 &r ) { x /= r.x; y /= r.y; z /= r.z; return *this; }

		//!< [NA
		void resetZero() { x = y = z = 0.0f; }

		//!< [xNg擾
		static const Vec3 &zero() {
			static Vec3 v0;
			return v0;
		}

		//!< ׂ(x,y)
		float lenSq2() const { return x * x + y * y; }
		//!< ׂ
		float lenSq() const { return x * x + y * y + z * z; }

		//!< (x,y)
		float len2() const { return sqrtf( lenSq2() ); }
		//!< 
		float len() const { return sqrtf( lenSq() ); }

		//!< K(x,ŷ)
		Vec3 &norm2() {
			float l = len2();
			if ( l <= 0.0f )
				return *this;
			x /= l; y /= l;
			return *this;
		}
		//!< K
		Vec3 &norm() {
			float l = len();
			if ( l <= 0.0f )
				return *this;
			x /= l; y /= l; z /= l;
			return *this;
		}

		//!< j󐳋K(x,ŷ)
		Vec3 normND2() const {
			float l = len2();
			if ( l <= 0.0f )
				return Vec3( 0.0f, 0.0f, z );
			return Vec3( x / l, y / l, z );
		}
		//!< j󐳋K
		Vec3 normND() const {
			float l = len();
			if ( l <= 0.0f )
				return Vec3();
			return Vec3( x / l, y / l, z / l );
		}

		//!< (x,ŷ)
		float dot2( const Vec3 &r ) { return x * r.x + y * r.y; }
		//!< 
		float dot3( const Vec3 &r ) { return x * r.x + y * r.y + z * r.z; }

		//!< O(x,ŷ)
		float cross2( const Vec3 &r ) const { return x * r.y - y * r.x; }
		//!< O(x,y,ẑ)
		Vec3 cross( const Vec3 &r ) const {
			return Vec3(
				y * r.z - z * r.y,
				z * r.x - x * r.z,
				x * r.y - y * r.x );
		}

		//!< Βl
		Vec3 abs() const {
			return Vec3( fabsf( x ), fabsf( y ), fabsf( z ) );
		}
	};

	/**
	* @brief xNg(4)
	*
	* (x,y,z,w)4xNgł
	*/
	class Vec4 {
	public:
		union {
			struct {float x, y, z, w;};
			float f[4];
		};

	public:
		Vec4(float x = 0.0f, float y = 0.0f, float z = 0.0f, float w = 0.0f) : x(x), y(y), z(z), w(w) {}
		Vec4( const Vec3 &v3, float w = 0.0f ) : x( v3.x ), y( v3.y ), z( v3.z ), w( w ) {}

		//!< +Zq
		Vec4 operator +(const Vec4 &r) const { return Vec4(x + r.x, y + r.y, z + r.z, w + r.w); }
		//!< -Zq
		Vec4 operator -(const Vec4 &r) const { return Vec4(x - r.x, y - r.y, z - r.z, w - r.w); }
		//!< *Zq
		Vec4 operator *( const float &r ) const { return Vec4( x * r, y * r, z * r, w * r ); }
		//!< *Zq
		Vec4 operator *(const Vec4 &r) const { return Vec4(x * r.x, y * r.y, z * r.z, w * r.w); }
		//!< /Zq
		Vec4 operator /( const float &r ) const { return Vec4( x / r, y / r, z / r, w / r ); }
		//!< /Zq
		Vec4 operator /(const Vec4 &r) const { return Vec4(x / r.x, y / r.y, z / r.z, w / r.w); }
		//!< +=Zq
		Vec4 &operator +=(const Vec4 &r) { x += r.x; y += r.y; z += r.z; w += r.w; return *this; }
		//!< -=Zq
		Vec4 &operator -=(const Vec4 &r) { x -= r.x; y -= r.y; z -= r.z; w -= r.w; return *this; }
		//!< *=Zq
		Vec4 &operator *=( const float &r ) { x *= r; y *= r; z *= r; w *= r; return *this; }
		//!< *=Zq
		Vec4 &operator *=(const Vec4 &r) { x *= r.x; y *= r.y; z *= r.z; w *= r.w; return *this; }
		//!< /=Zq
		Vec4 &operator /=( const float &r ) { x /= r; y /= r; z /= r; w /= r; return *this; }
		//!< /=Zq
		Vec4 &operator /=(const Vec4 &r) { x /= r.x; y /= r.y; z /= r.z; w /= r.w; return *this; }
		//!< =Zq
		operator Vec3() const {
			return Vec3( x, y, z );
		}

		//!< [NA
		void resetZero() { x = y = z = w = 0.0f; }

		//!< [xNg擾
		static const Vec4 &zero() {
			static Vec4 v0;
			return v0;
		}

		//!< ׂ(x,y)
		float lenSq2() const { return x * x + y * y; }
		//!< ׂ(x,y,z)
		float lenSq3() const { return x * x + y * y + z * z; }
		//!< ׂ
		float lenSq() const { return x * x + y * y + z * z + w * w; }

		//!< (x,y)
		float len2() const { return sqrtf(lenSq2()); }
		//!< (x,y,z)
		float len3() const { return sqrtf(lenSq3()); }
		//!< 
		float len() const { return sqrtf(lenSq()); }

		//!< K(x,ŷ)
		Vec4 &norm2() {
			float l = len2();
			if ( l <= 0.0f )
				return *this;
			x /= l; y /= l;
			return *this;
		}
		//!< K(x,y,ẑ)
		Vec4 &norm3() {
			float l = len3();
			if ( l <= 0.0f )
				return *this;
			x /= l; y /= l; z /= l;
			return *this;
		}
		//!< K
		Vec4 &norm() {
			float l = len();
			if ( l <= 0.0f )
				return *this;
			x /= l; y /= l; z /= l; w /= l;
			return *this;
		}

		//!< j󐳋K(x,ŷ)
		Vec4 normND2() const {
			float l = len2();
			if ( l <= 0.0f )
				return Vec4( 0.0f, 0.0f, z, w );
			return Vec4( x / l, y / l, z, w );
		}
		//!< j󐳋K(x,y,ẑ)
		Vec4 normND3() const {
			float l = len3();
			if ( l <= 0.0f )
				return Vec4( 0.0f, 0.0f, 0.0f, w );
			return Vec4(x / l, y / l, z / l, w);
		}
		//!< j󐳋K
		Vec4 normND() const {
			float l = len();
			if ( l <= 0.0f )
				return Vec4();
			return Vec4(x / l, y / l, z / l, w / l);
		}

		//!< (x,ŷ)
		float dot2(const Vec4 &r) { return x * r.x + y * r.y; }
		//!< (x,y,ẑ)
		float dot3(const Vec4 &r) { return x * r.x + y * r.y + z * r.z; }
		//!< 
		float dot(const Vec4 &r) { return x * r.x + y * r.y + z * r.z + w * r.w; }

		//!< O(x,ŷ)
		float cross2(const Vec4 &r) const { return x * r.y - y * r.x; }
		//!< O(x,y,ẑ)
		Vec4 cross3(const Vec4 &r) const {
			return Vec4(
				y * r.z - z * r.y,
				z * r.x - x * r.z,
				x * r.y - y * r.x);
		}
	};


	// 4x4s
	class Mat4x4 {
	public:
		union {
			float f[16];
			float rc[4][4];
			struct { float m11, m12, m13, m14, m21, m22, m23, m24, m31, m32, m33, m34, m41, m42, m43, m44; };
			struct { Vec4 v[4]; };
		};

		void _swap(float &a, float &b) {float t = a; a = b; b = t;}

	public:
		Mat4x4() {}
		Mat4x4(const float _f[16]) { memcpy(f, _f, sizeof(float) * 16); }
		Mat4x4(const Vec4 &r1, const Vec4 &r2, const Vec4 &r3, const Vec4 &r4) { v[0] = r1; v[1] = r2; v[2] = r3; v[3] = r4; }
		Mat4x4(
			float m11, float m12, float m13, float m14,
			float m21, float m22, float m23, float m24,
			float m31, float m32, float m33, float m34,
			float m41, float m42, float m43, float m44) :
			m11(m11), m12(m12), m13(m13), m14(m14),
			m21(m21), m22(m22), m23(m23), m24(m24),
			m31(m31), m32(m32), m33(m33), m34(m34),
			m41(m41), m42(m42), m43(m43), m44(m44) {}

		// RCANZX
		float& operator()(uint32_t _r, uint32_t _c) {
			return rc[_r][_c];
		}

		Mat4x4 operator *(const Mat4x4 &r) const {
			Mat4x4 res(
				m11 * r.m11 + m12 * r.m21 + m13 * r.m31 + m14 * r.m41,
				m11 * r.m12 + m12 * r.m22 + m13 * r.m32 + m14 * r.m42,
				m11 * r.m13 + m12 * r.m23 + m13 * r.m33 + m14 * r.m43,
				m11 * r.m14 + m12 * r.m24 + m13 * r.m34 + m14 * r.m44,

				m21 * r.m11 + m22 * r.m21 + m23 * r.m31 + m24 * r.m41,
				m21 * r.m12 + m22 * r.m22 + m23 * r.m32 + m24 * r.m42,
				m21 * r.m13 + m22 * r.m23 + m23 * r.m33 + m24 * r.m43,
				m21 * r.m14 + m22 * r.m24 + m23 * r.m34 + m24 * r.m44,

				m31 * r.m11 + m32 * r.m21 + m33 * r.m31 + m34 * r.m41,
				m31 * r.m12 + m32 * r.m22 + m33 * r.m32 + m34 * r.m42,
				m31 * r.m13 + m32 * r.m23 + m33 * r.m33 + m34 * r.m43,
				m31 * r.m14 + m32 * r.m24 + m33 * r.m34 + m34 * r.m44,

				m41 * r.m11 + m42 * r.m21 + m43 * r.m31 + m44 * r.m41,
				m41 * r.m12 + m42 * r.m22 + m43 * r.m32 + m44 * r.m42,
				m41 * r.m13 + m42 * r.m23 + m43 * r.m33 + m44 * r.m43,
				m41 * r.m14 + m42 * r.m24 + m43 * r.m34 + m44 * r.m44
			);

			return res;
		}

		Mat4x4 operator *(const float &r) const {
			Mat4x4 res(
				m11 * r, m12 * r, m13 * r, m14 * r,
				m21 * r, m22 * r, m23 * r, m24 * r,
				m31 * r, m32 * r, m33 * r, m34 * r,
				m41 * r, m42 * r, m43 * r, m44 * r
			);

			return res;
		}

		friend Vec3 operator *( const Vec3 &l, const Mat4x4 &m ) {
			return Vec3(
				l.x * m.m11 + l.y * m.m21 + l.z * m.m31,
				l.x * m.m12 + l.y * m.m22 + l.z * m.m32,
				l.x * m.m13 + l.y * m.m23 + l.z * m.m33
				);
		}

		friend Vec4 operator *( const Vec4 &l, const Mat4x4 &m ) {
			return Vec4(
				l.x * m.m11 + l.y * m.m21 + l.z * m.m31 + l.w * m.m41,
				l.x * m.m12 + l.y * m.m22 + l.z * m.m32 + l.w * m.m42,
				l.x * m.m13 + l.y * m.m23 + l.z * m.m33 + l.w * m.m43,
				l.x * m.m14 + l.y * m.m24 + l.z * m.m34 + l.w * m.m44
				);
		}

		// *=
		Mat4x4 &operator *=(const Mat4x4 &r) {
			*this = *this * r;
			return *this;
		}
		Mat4x4 &operator *=(const float &r) {
			*this = *this * r;
			return *this;
		}

		// Pʍs擾
		static const Mat4x4 &iden() {
			static const Mat4x4 i(
				1.0f, 0.0f, 0.0f, 0.0f,
				0.0f, 1.0f, 0.0f, 0.0f,
				0.0f, 0.0f, 1.0f, 0.0f,
				0.0f, 0.0f, 0.0f, 1.0f);
			return i;
		}

		// Pʍs
		Mat4x4 &resetIden() {
			memcpy(v, &iden(), sizeof(v));
			return *this;
		}

		// [s擾
		static const Mat4x4 &zero() {
			static const Mat4x4 m0(
				0.0f, 0.0f, 0.0f, 0.0f,
				0.0f, 0.0f, 0.0f, 0.0f,
				0.0f, 0.0f, 0.0f, 0.0f,
				0.0f, 0.0f, 0.0f, 0.0f);
			return m0;
		}

		// [
		Mat4x4 &resetZero() {
			memcpy(v, &zero(), sizeof(v));
			return *this;
		}

		// ]u
		Mat4x4 &trans() {
			_swap(m12, m21);
			_swap(m13, m31);
			_swap(m14, m41);
			_swap(m23, m32);
			_swap(m24, m42);
			_swap(m34, m43);
			return *this;
		}

		// j]u
		Mat4x4 transND() const {
			return Mat4x4(
				m11, m21, m31, m41,
				m12, m22, m32, m42,
				m13, m23, m33, m43,
				m14, m24, m34, m44
			);
		}

		// s񎮎擾
		float det() const {
			return m11 * m22 * m33 * m44 + m11 * m23 * m34 * m42 + m11 * m24 * m32 * m43
				+  m12 * m21 * m34 * m43 + m12 * m23 * m31 * m44 + m12 * m24 * m33 * m41
				+  m13 * m21 * m32 * m44 + m13 * m22 * m34 * m41 + m13 * m24 * m31 * m42
				+  m14 * m21 * m33 * m42 + m14 * m22 * m31 * m43 + m14 * m23 * m32 * m41
				-  m11 * m22 * m34 * m43 - m11 * m23 * m32 * m44 - m11 * m24 * m33 * m42
				-  m12 * m21 * m33 * m44 - m12 * m23 * m34 * m41 - m12 * m24 * m31 * m43
				-  m13 * m21 * m34 * m42 - m13 * m22 * m31 * m44 - m13 * m24 * m32 * m41
				-  m14 * m21 * m32 * m43 - m14 * m22 * m33 * m41 - m14 * m23 * m31 * m42;
		}

		// ts
		Mat4x4 &inv() {
			float d = det();
			if (d == 0.0f)
				return *this;
			float b11 = m22 * m33 * m44 + m23 * m34 * m42 + m24 * m32 * m43 - m22 * m34 * m43 - m23 * m32 * m44 - m24 * m33 * m42;
			float b12 = m12 * m34 * m43 + m13 * m32 * m44 + m14 * m33 * m42 - m12 * m33 * m44 - m13 * m34 * m42 - m14 * m32 * m43;
			float b13 = m12 * m23 * m44 + m13 * m24 * m42 + m14 * m22 * m43 - m12 * m24 * m43 - m13 * m22 * m44 - m14 * m23 * m42;
			float b14 = m12 * m24 * m33 + m13 * m22 * m34 + m14 * m23 * m32 - m12 * m23 * m34 - m13 * m24 * m32 - m14 * m22 * m33;
			float b21 = m21 * m34 * m43 + m23 * m31 * m44 + m24 * m33 * m41 - m21 * m33 * m44 - m23 * m34 * m41 - m24 * m31 * m43;
			float b22 = m11 * m33 * m44 + m13 * m34 * m41 + m14 * m31 * m43 - m11 * m34 * m43 - m13 * m31 * m44 - m14 * m33 * m41;
			float b23 = m11 * m24 * m43 + m13 * m21 * m44 + m14 * m23 * m41 - m11 * m23 * m44 - m13 * m24 * m41 - m14 * m21 * m43;
			float b24 = m11 * m23 * m34 + m13 * m24 * m31 + m14 * m21 * m33 - m11 * m24 * m33 - m13 * m21 * m34 - m14 * m23 * m31;
			float b31 = m21 * m32 * m44 + m22 * m34 * m41 + m24 * m31 * m42 - m21 * m34 * m42 - m22 * m31 * m44 - m24 * m32 * m41;
			float b32 = m11 * m34 * m42 + m12 * m31 * m44 + m14 * m32 * m41 - m11 * m32 * m44 - m12 * m34 * m41 - m14 * m31 * m42;
			float b33 = m11 * m22 * m44 + m12 * m24 * m41 + m14 * m21 * m42 - m11 * m24 * m42 - m12 * m21 * m44 - m14 * m22 * m41;
			float b34 = m11 * m24 * m32 + m12 * m21 * m34 + m14 * m22 * m31 - m11 * m22 * m34 - m12 * m24 * m31 - m14 * m21 * m32;
			float b41 = m21 * m33 * m42 + m22 * m31 * m43 + m23 * m32 * m41 - m21 * m32 * m43 - m22 * m33 * m41 - m23 * m31 * m42;
			float b42 = m11 * m32 * m43 + m12 * m33 * m41 + m13 * m31 * m42 - m11 * m33 * m42 - m12 * m31 * m43 - m13 * m32 * m41;
			float b43 = m11 * m23 * m42 + m12 * m21 * m43 + m13 * m22 * m41 - m11 * m22 * m43 - m12 * m23 * m41 - m13 * m21 * m42;
			float b44 = m11 * m22 * m33 + m12 * m23 * m31 + m13 * m21 * m32 - m11 * m23 * m32 - m12 * m21 * m33 - m13 * m22 * m31;

			m11 = b11/d; m12 = b12/d; m13 = b13/d; m14 = b14/d;
			m21 = b21/d; m22 = b22/d; m23 = b23/d; m24 = b24/d;
			m31 = b31/d; m32 = b32/d; m33 = b33/d; m34 = b34/d;
			m41 = b41/d; m42 = b42/d; m43 = b43/d; m44 = b44/d;

			return *this;
		}

		// jts
		Mat4x4 invND() const {
			float d = det();
			if (d == 0.0f)
				return *this;
			Mat4x4 out (
				m22 * m33 * m44 + m23 * m34 * m42 + m24 * m32 * m43 - m22 * m34 * m43 - m23 * m32 * m44 - m24 * m33 * m42,
				m12 * m34 * m43 + m13 * m32 * m44 + m14 * m33 * m42 - m12 * m33 * m44 - m13 * m34 * m42 - m14 * m32 * m43,
				m12 * m23 * m44 + m13 * m24 * m42 + m14 * m22 * m43 - m12 * m24 * m43 - m13 * m22 * m44 - m14 * m23 * m42,
				m12 * m24 * m33 + m13 * m22 * m34 + m14 * m23 * m32 - m12 * m23 * m34 - m13 * m24 * m32 - m14 * m22 * m33,

				m21 * m34 * m43 + m23 * m31 * m44 + m24 * m33 * m41 - m21 * m33 * m44 - m23 * m34 * m41 - m24 * m31 * m43,
				m11 * m33 * m44 + m13 * m34 * m41 + m14 * m31 * m43 - m11 * m34 * m43 - m13 * m31 * m44 - m14 * m33 * m41,
				m11 * m24 * m43 + m13 * m21 * m44 + m14 * m23 * m41 - m11 * m23 * m44 - m13 * m24 * m41 - m14 * m21 * m43,
				m11 * m23 * m34 + m13 * m24 * m31 + m14 * m21 * m33 - m11 * m24 * m33 - m13 * m21 * m34 - m14 * m23 * m31,

				m21 * m32 * m44 + m22 * m34 * m41 + m24 * m31 * m42 - m21 * m34 * m42 - m22 * m31 * m44 - m24 * m32 * m41,
				m11 * m34 * m42 + m12 * m31 * m44 + m14 * m32 * m41 - m11 * m32 * m44 - m12 * m34 * m41 - m14 * m31 * m42,
				m11 * m22 * m44 + m12 * m24 * m41 + m14 * m21 * m42 - m11 * m24 * m42 - m12 * m21 * m44 - m14 * m22 * m41,
				m11 * m24 * m32 + m12 * m21 * m34 + m14 * m22 * m31 - m11 * m22 * m34 - m12 * m24 * m31 - m14 * m21 * m32,

				m21 * m33 * m42 + m22 * m31 * m43 + m23 * m32 * m41 - m21 * m32 * m43 - m22 * m33 * m41 - m23 * m31 * m42,
				m11 * m32 * m43 + m12 * m33 * m41 + m13 * m31 * m42 - m11 * m33 * m42 - m12 * m31 * m43 - m13 * m32 * m41,
				m11 * m23 * m42 + m12 * m21 * m43 + m13 * m22 * m41 - m11 * m22 * m43 - m12 * m23 * m41 - m13 * m21 * m42,
				m11 * m22 * m33 + m12 * m23 * m31 + m13 * m21 * m32 - m11 * m23 * m32 - m12 * m21 * m33 - m13 * m22 * m31
			);
			
			return out * (1.0f / d);
		}
	};
}

#endif
