#include "oxstlparser.h"
#include <stdint.h>
#include <fstream>


namespace {
	using namespace OX;

#pragma pack( push, 1 )
	struct STLTriangle {
		Vec3 normal_;	// @
		Vec3 p_[ 3 ];	// _
		uint16_t reserved_;	// \̈
	};

	struct STLHeader {
		char header_[ 80 ];	// 80oCg̕
		uint32_t num_;		// Op`̖
		STLTriangle triangle[ 1 ];	// Op`
	};
#pragma pack( pop )
}


namespace OX {
	STLParser::STLParser() {

	}

	STLParser::~STLParser() {

	}

	bool STLParser::parse( const char* stlData, size_t size, bool isBynary ) {
		if ( stlData == 0 || size == 0 )
			return false;
		if ( isBynary == false )
			return false;	// TODO: eLXg͎芸͈Ȃ

		const STLHeader *header = (STLHeader*)stlData;
		uint32_t count = 0;
		for ( size_t i = 0; i < header->num_; ++i ) {
			for ( int j = 0; j < 3; ++j ) {
				Vertex v;
				v.p_ = header->triangle[ i ].p_[ j ];
				v.n_ = header->triangle[ i ].normal_;
				vertices_.push_back( v );

				indices_.push_back( count );
				count++;
			}
		}

		return true;
	}

	bool STLParser::parseFromFile( const char* filePath, bool isBynary ) {
		if ( isBynary == false )
			return false;	// TODO: eLXg͎芸͈Ȃ

		std::ifstream ifs;
		ifs.open( filePath, std::ios::in | std::ios::binary );
		if ( !ifs )
			return false;

		ifs.seekg( 0, std::ios_base::end );
		std::streamoff sz = ifs.tellg();
		ifs.seekg( 0, std::ios_base::beg );

		char *mem = new char[ (unsigned int)sz ];
		ifs.read( mem, sz );
		ifs.close();

		bool res = parse( mem, (size_t)sz, isBynary );
		delete mem;

		return res;
	}

	const std::vector< STLParser::Vertex > &STLParser::getVertices() const {
		return vertices_;
	}

	size_t STLParser::getVerticesSize() const {
		return vertices_.size() * sizeof( Vertex );
	}

	const std::vector< uint32_t > &STLParser::getIndices() const {
		return indices_;
	}

	AABB STLParser::calcAABB() const {
		AABB aabb = AABB( Vec3() );
		if ( vertices_.size() == 0 ) {
			return aabb;
		}
		for ( size_t i = 0; i < vertices_.size(); ++i ) {
			aabb.updateRegion( vertices_[ i ].p_ );
		}
		return aabb;
	}

}